# Resthys - Une interface REST partielle pour Noethys
# Copyright (C) 2018  Cliss XXI <tech@cliss21.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from decimal import Decimal

from django.utils.text import slugify

from resthys.base import models

logger = logging.getLogger(__name__)


def int_cents_from_any_float(a):
    return int(Decimal(a).quantize(Decimal('0.01')).shift(2))


def get_typed_montant(o):
    # The typed amount.
    try:
        return int_cents_from_any_float(o.montant)
    except Exception as e:
        logger.error(e)
        return 'E'


def get_prestation_montant(p):
    return get_typed_montant(p)


def get_ventilation_montant(v):
    return get_typed_montant(v)


"""
We recompute the sum of the breakdowns amounts associated to the prestation,
limited by the prestation cost.
The main code doing this is here:
    https://github.com/Noethys/Noethys/blob/master/noethys/Ol/OL_Factures.py#L254
"""


def get_prestation_regle(p):
    try:
        return min(
            get_prestation_montant(p),
            sum(
                [
                    get_ventilation_montant(v)
                    for v in models.Ventilation.objects.filter(prestation=p.id)
                ]
            ),
        )
    except Exception as e:
        logger.error(e)
        return 'E'


"""
The computed due amount.
"""


def get_prestation_solde(p):
    try:
        return max(0, get_prestation_montant(p) - get_prestation_regle(p))
    except Exception as e:
        logger.error(e)
        return 'E'


def get_facture_montant(f):
    try:
        f_amount = int_cents_from_any_float(f.total)
        p_amount = sum([
            get_prestation_montant(p)
            for p in f.prestations_set.all()
            ])
        if f_amount != p_amount:
            raise ValueError(
                'Montant calculé différent du montant affiché '
                'pour la facture n° {}.'.format(f.numero)
            )
        return f_amount
    except Exception as e:
        logger.error(e)
        return 'E'


def get_facture_regle(f):
    try:
        return sum([get_prestation_regle(p) for p in f.prestations_set.all()])
    except Exception as e:
        logger.error(e)
        return 'E'


def get_facture_solde(f):
    try:
        return sum([get_prestation_solde(p) for p in f.prestations_set.all()])
    except Exception as e:
        logger.error(e)
        return 'E'


"""
We recompute the displayed names in invoice.
The main code doing this is here:
    https://github.com/Noethys/Noethys/blob/master/noethys/Utils/UTILS_Titulaires.py#L114
"""


def get_titulaires_prettyprinted(f):
    titus = f.rattachements_set.filter(titulaire=True)
    if titus.count() == 2 and titus[0].individu.nom == titus[1].individu.nom:
        # special case
        return "{} et {} {}".format(
            titus[0].individu.prenom,
            titus[1].individu.prenom,
            titus[0].individu.nom,
        )
    # normal case
    titus = ['{} {}'.format(t.individu.prenom, t.individu.nom) for t in titus]
    if len(titus) == 0:
        return ""
    elif len(titus) == 1:
        return titus[0]
    else:
        last = titus.pop()
        return '{} et {}'.format(', '.join(titus), last)


def match_facture_names(f, name):
    def slug_and_split(s):
        return sorted(slugify(s).split('-'))

    normalized_query = slug_and_split(name)
    normalized_names = [
        slug_and_split("{} {}".format(p.individu.nom, p.individu.prenom))
        for p in f.compte_payeur.famille.rattachements_set.filter(
            titulaire=True
        )
    ]
    normalized_family = [
        slug_and_split(get_titulaires_prettyprinted(f.compte_payeur.famille))
    ]
    return normalized_query in normalized_names + normalized_family


"""
We recompute individu's sexe.
The main interresting code is here:
    https://github.com/Noethys/Noethys/blob/master/noethys/Data/DATA_Civilites.py
"""


def get_sexe(i):
    if i.idcivilite in [2, 3, 5]:
        return 'F'
    elif i.idcivilite in [1, 4]:
        return 'M'
    return ''


"""
We retrieve the address of compte_payeur.
The main code doing this is here:
    https://github.com/Noethys/Noethys/blob/master/noethys/Utils/UTILS_Titulaires.py#L143
"""


def get_compte_payeur_adresse(obj):
    i = obj.famille.rattachements_set.filter(titulaire=True)[0].individu
    if i.adresse_auto:
        i = obj.famille.rattachements_set.get(pk=i)
    return {
        'rue': i.rue_resid,
        'code_postal': i.cp_resid,
        'ville': i.ville_resid,
    }


"""
We retrieve the email address of compte_payeur.
"""


def get_compte_payeur_email(obj):
    tits = [
            r.individu.email
            for r in obj.famille.rattachements_set.filter(titulaire=True)
            if r.individu.email is not None
            ]
    return tits[0] if len(tits) > 0 else None


"""
We retrieve other unpaid invoices with same compte_payeur.
"""


def get_autres_factures(f):
    cp = f.compte_payeur
    lf = models.Factures.objects.filter(compte_payeur=cp).exclude(pk=f.id)
    lf = [(f, get_facture_solde(f)) for f in lf]
    return [{'numero': f.numero, 'solde': s} for f, s in lf if s > 0]
