# Resthys - Une interface REST partielle pour Noethys
# Copyright (C) 2018  Cliss XXI <tech@cliss21.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from collections import OrderedDict
from django.db import connections
from django.core.management.base import CommandError

from django.core.management.commands.inspectdb import (
    Command as InspectDBCommand
)


class Command(InspectDBCommand):
    def handle(self, **options):
        if options.get('database') != 'noethys':
            self.handle_inspection = super().handle_inspection
        try:
            options['table_name_filter'] = lambda x: x != 'django_migrations'
            for line in self.handle_inspection(options):
                self.stdout.write("%s\n" % line)
        except NotImplementedError:
            raise CommandError(
                "Database inspection isn't supported for the currently "
                "selected database backend."
            )

    def normalize_col_name(
        self, table_name, col_name, used_column_names, is_relation, is_primary
    ):
        new_name, field_params, field_notes = super().normalize_col_name(
            col_name, used_column_names, is_relation
        )
        in_conflicts = {('factures', 'prestations'): 'prestations_list'}
        if is_primary:
            new_name = 'id'
        elif is_relation and new_name.startswith('id') and len(new_name) > 2:
            new_name = new_name[2:]
        elif new_name == 'mail' or new_name == 'travail_mail':
            new_name = new_name.replace('mail', 'email')
            field_params['db_column'] = col_name
        elif (table_name, col_name) in in_conflicts:
            new_name = in_conflicts[(table_name, col_name)]
            field_params['db_column'] = col_name

        return new_name, field_params, field_notes

    def get_field_type(self, conn, table_name, row):
        field_type, field_params, field_notes = super().get_field_type(
            conn, table_name, row
        )

        # Blobs
        if row[0] == 'logo':
            field_type = 'BinaryField'

        # Noethys split datetime between time and date columns
        if field_type == 'DateField' and 'heure' in row[0]:
            field_type = 'TimeField'

        # Email fields
        if field_type == 'CharField' and row[0] in ['mail', 'travail_mail']:
            field_type = 'EmailField'
        return field_type, field_params, field_notes

    def handle_inspection(self, options):
        # on delete matrix for foreign key
        on_delete = {
            #       ('consommations', 'IDindividu'): 'CASCADE',
        }
        connection = connections[options['database']]
        # 'table_name_filter' is a stealth option
        table_name_filter = options.get('table_name_filter')

        def table2model(table_name):
            return re.sub(r'[^a-zA-Z0-9]', '', table_name.title())

        def strip_prefix(s):
            return s[1:] if s.startswith("u'") else s

        def guess_relations(tables, connection):
            """
            Given a column and a table where it appears, we need to know:
            * if it refers to an existant modele primary key
              (not an hardcoded ID);
            * which model (because multiple models may have the same
              primary_key name).
            """
            relations = {
                # attended format:
                # dbcol: [table1, table2, ...]
                # => dbcol exists as pk in [table1, table2, ...]
                #
                # and manualy disambigued relations:
                # (dbcol,table): table_dest
                # => dbcol in table in known to refers to table_dest
                ('IDmodele', 'documents_objets'): 'documents_modeles'
            }
            for table_name in tables:
                try:
                    tdesc = connection.introspection.get_table_description(
                        cursor, table_name
                    )
                    primary_key_column = tdesc[0].name
                    if primary_key_column in relations:
                        relations[primary_key_column].append(table_name)
                    else:
                        relations[primary_key_column] = [table_name]
                except Exception as e:
                    continue
            return relations

        with connection.cursor() as cursor:
            yield "# This is an auto-generated Django model module."
            yield 'from %s import models' % self.db_module
            known_models = []
            tables_to_introspect = options[
                'table'
            ] or connection.introspection.table_names(cursor)
            relations = guess_relations(tables_to_introspect, connection)

            for table_name in tables_to_introspect:
                if table_name_filter is not None and callable(
                    table_name_filter
                ):
                    if not table_name_filter(table_name):
                        continue
                try:
                    tdesc = connection.introspection.get_table_description(
                        cursor, table_name
                    )
                    primary_key_column = tdesc[0].name
                except Exception as e:
                    yield "# Unable to inspect table '%s'" % table_name
                    yield "# The error was: %s" % e
                    continue

                yield ''
                yield ''
                yield 'class %s(models.Model):' % table2model(table_name)
                known_models.append(table2model(table_name))
                # Holds column names used in the table so far
                used_column_names = []
                # Maps column names to names of model fields
                column_to_field_name = {}
                for row in tdesc:
                    # Holds Field notes, to be displayed in a Python comment.
                    comment_notes = []
                    extra_params = (
                        OrderedDict()
                    )  # Holds Field parameters such as 'db_column'.
                    column_name = row[0]
                    is_relation = column_name in relations
                    is_primary = column_name == primary_key_column

                    att_name, params, notes = self.normalize_col_name(
                        table_name,
                        column_name,
                        used_column_names,
                        is_relation,
                        is_primary,
                    )
                    extra_params.update(params)
                    comment_notes.extend(notes)

                    used_column_names.append(att_name)
                    column_to_field_name[column_name] = att_name

                    if is_primary:
                        extra_params['primary_key'] = True

                    if is_relation and column_name != primary_key_column:
                        if (column_name, table_name) in relations and relations[
                            (column_name, table_name)
                        ] == table_name:
                            rel_to = "self"
                        elif (column_name, table_name) in relations:
                            rel_to = table2model(
                                relations[(column_name, table_name)]
                            )
                        else:
                            dest_models = relations[column_name]
                            if len(dest_models) > 1:
                                comment_notes.append(
                                    'Ambiguous foreign key. You should '
                                    'precise it in the relation dict.'
                                )
                            rel_to = table2model(dest_models[-1])

                        if rel_to in known_models:
                            field_type = 'ForeignKey(%s' % rel_to
                        else:
                            field_type = "ForeignKey('%s'" % rel_to
                    else:
                        # Calling `get_field_type` to get the field type
                        # string and any additional parameters and notes.
                        field_type, f_params, f_notes = self.get_field_type(
                            connection, table_name, row
                        )
                        extra_params.update(f_params)
                        comment_notes.extend(f_notes)

                        field_type += '('

                    # Don't output 'id = meta.AutoField(primary_key=True)',
                    # because that's assumed if it doesn't exist.
                    if (
                        att_name == 'id'
                        and 'primary_key' in extra_params
                        and extra_params['primary_key'] is True
                    ):
                        if field_type == 'AutoField(' and extra_params == {
                            'primary_key': True
                        }:
                            continue
                        elif (
                            field_type == 'IntegerField('
                            and not connection.features.can_introspect_autofield
                        ):
                            field_type = 'AutoField('
                            comment_notes.append(
                                'Guessed AutoField, since DB connection '
                                'cannot introspect it.'
                            )

                    # Add 'null' and 'blank', if the 'null_ok' flag was present
                    # in the table description.
                    if row[6]:  # If it's NULL...
                        if field_type == 'BooleanField(':
                            field_type = 'NullBooleanField('
                        elif (
                            'primary_key' not in extra_params
                            or extra_params['primary_key'] is not True
                        ):
                            extra_params['blank'] = True
                            extra_params['null'] = True

                    field_desc = '%s = %s%s' % (
                        att_name,
                        # Custom fields will have a dotted path
                        '' if '.' in field_type else 'models.',
                        field_type,
                    )
                    if field_type.startswith('ForeignKey('):
                        if (table_name, column_name) in on_delete:
                            field_desc += ', on_delete=models.{}'.format(
                                on_delete[(table_name, column_name)]
                            )
                        else:
                            field_desc += ', on_delete=models.SET_NULL'

                    if extra_params:
                        if not field_desc.endswith('('):
                            field_desc += ', '
                        field_desc += ', '.join(
                            '%s=%s' % (k, strip_prefix(repr(v)))
                            for k, v in extra_params.items()
                        )
                    field_desc += ')'
                    if comment_notes:
                        field_desc += '  # ' + ' '.join(comment_notes)
                    if len(field_desc) > 72:
                        field_desc += '  # noqa'
                    yield '    %s' % field_desc
                if 'nom' in used_column_names:
                    yield ''
                    yield '    def __str__(self):'
                    yield '        return self.nom'
                for meta_line in self.get_meta(
                    table_name, {}, column_to_field_name
                ):
                    yield meta_line
